%{
/*
 *  ippl.y - functions parsing the configuration file
 *
 *  Copyright (C) 1998, 1999 Etienne Bernard
 *  *Small* modifications for UDP by Hugo Haas, 1999
 *  Now resolves IP addresses at parsing time whenever possible,
 *    Etienne Bernard 1999
 *  Keeps hostname for refreshing. Added "expire" rule. Hugo Haas 1999
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>

#include "configuration.h"
#include "filter.h"

void yyerror(char *);
void print_error(char *, int);
int yylex(void);

char *proto;

extern struct filter_entry *filter;
extern unsigned short log_protocols;
extern unsigned int dns_expire;
extern int line_number;

%}

%union {
  long longval;
  char * stringval;
  struct portrange_struct rangeval;
  struct hostdesc_struct hostval;
  union loginfo_union loginfoval;
  struct filter_entry * entryval;
  struct {
    long protocol;
    union loginfo_union loginfoval;
  } protocolval;
}

%token<stringval> IP HOSTMASK IDENT
%token<longval>    NUMBER

%token RUN EXPIRE LOG IGNORE FROM TYPE PORT
%token ICMP TCP UDP ALL

%token RANGE SLASH EOL

%type<entryval> Rule
%type<longval> Action
%type<rangeval> Range
%type<hostval> From
%type<protocolval> Protocol
%type<rangeval> Port
%type<loginfoval> Type

%start Input
%%

Input:
          /* Empty */
        | Input Line
        ;

Line:
          EOL
        | RUN LogList EOL
        | EXPIRE NUMBER
          { dns_expire = $2; }
        | Rule EOL
          { $1->next = filter; filter = $1; }
        ;

LogList:
          LogItem
        | LogList LogItem
        ;

LogItem:
          ICMP
          { log_protocols |= RUN_ICMP; }
        | TCP
          { log_protocols |= RUN_TCP;  }
        | UDP
          { log_protocols |= RUN_UDP;  }
        | ALL
          { log_protocols |= (RUN_TCP & RUN_ICMP & RUN_UDP); }
        | error
          { print_error("unknown protocol.", line_number); yyclearin; }
        ;

Rule:
          Action Protocol From
          {
            $$ = (struct filter_entry *)malloc(sizeof(struct filter_entry));
            $$->log = $1;
            $$->protocol = $2.protocol;
            $$->loginfo = $2.loginfoval;
            $$->hostdesc = $3;
          }
        | error EOL
          { print_error("ignoring invalid rule.", line_number-1); }
        ;

Action:
          LOG     { $$ = 1; }
        | IGNORE  { $$ = 0; }
        ;

Protocol:
          ICMP Type
          { $$.protocol = ICMP_PROTO; $$.loginfoval = $2; }
        | TCP Port
          { $$.protocol = TCP_PROTO; $$.loginfoval = $2; proto = "tcp"; }
        | UDP Port
          { $$.protocol = UDP_PROTO; $$.loginfoval = $2; proto = "udp"; }
        ;

Type:
          { $$.icmptype = NO_ICMP_TYPE; }
        | TYPE NUMBER { $$.icmptype = $2; }
        ;

Port:
          { $$.range_min = NOT_DEFINED;
            $$.range_max = NOT_DEFINED; }
       |  PORT Range
          { $$ = $2; }
       ;

Range:
          NUMBER               { $$.range_min = $$.range_max = $1; }
	| IDENT                {
                                 struct servent *s = getservbyname($1, proto);
                                 if (s)
                                   $$.range_min = $$.range_max = ntohs(s->s_port);
                                 else {
                                   $$.range_min = $$.range_max = NOT_DEFINED;
                                   print_error("unknown service.", line_number);
                                   YYERROR;
                                 }
                                 free($1);
                               }
        | RANGE NUMBER         { $$.range_min = NOT_DEFINED;
                                 $$.range_max = $2;                    }
        | NUMBER RANGE         { $$.range_min = $1;
                                 $$.range_max = NOT_DEFINED;           }
        | NUMBER RANGE NUMBER  { $$.range_min = $1; $$.range_max = $3; }
        ;

From:
        
          {
            $$.address = $$.mask = 0;
            $$.hostname = NULL;
            $$.hostmask = NULL;
          }
        | FROM IP
            {
              struct in_addr inp;
              inet_aton($2, &inp);
              $$.address = inp.s_addr;
              $$.mask = 0xffffffff;
              $$.hostname = NULL;
              $$.hostmask = NULL;
              free($2);
            }
        | FROM IP SLASH IP
            {
              struct in_addr inp;
              inet_aton($2, &inp);
              $$.address = inp.s_addr;
              inet_aton($4, &inp);
              $$.mask = inp.s_addr;
              $$.hostname = NULL;              
              $$.hostmask = NULL;
              free($2); free($4);
            }
        | FROM HOSTMASK
            {
              if (strchr($2, '*')) {
                $$.hostmask = $2;
                $$.hostname = NULL;
                $$.address = 0;
                $$.mask = 0;
              } else {
                struct hostent * host;
                if ((host = gethostbyname((const char *)$2)) == 0) {
                  $$.address = 0;
                  $$.mask = 0;
                  print_error("cannot resolve hostname.", line_number);
                } else {
                  memcpy(&($$.address), host->h_addr, host->h_length);
                  $$.mask = 0xffffffff;
                };
                $$.hostname = $2;
                $$.hostmask = NULL;
              }
            }
        | FROM IDENT
            {
              struct hostent * host;
              if ((host = gethostbyname((const char *)$2)) == 0) {
                $$.address = 0;
                $$.mask = 0;
                print_error("cannot resolve hostname.", line_number);
              } else {
                memcpy(&($$.address), host->h_addr, host->h_length);
                $$.mask = 0xffffffff;
              };
              $$.hostname = $2;
              $$.hostmask = NULL;
            }
        ;

%%

void yyerror(char *s) {
  return;
}

void print_error(char *s, int l) {
  printf("Parse error line %d: %s\n",l,s);

}
